## ----loadinglibraries, include=FALSE, echo=FALSE, eval=T----------------------
library(Ribostan)

## -----------------------------------------------------------------------------
#first, let's load up some test data
anno_file <- here::here('test.gc32.gtf')
if(!file.exists(anno_file)){
  library(AnnotationHub)
  ah <- AnnotationHub()
  gencode32 <- ah[['AH75191']]
   seqlevels(gencode32)<-'chr22'
  rtracklayer::export(gencode32, anno_file)
}
fafile <- here::here('chr22.fa')
library(BSgenome.Hsapiens.UCSC.hg38)
if(!file.exists(fafile)){
  seq <- Biostrings::DNAStringSet(BSgenome.Hsapiens.UCSC.hg38[['chr22']])
  names(seq) <- 'chr22'
 Biostrings::writeXStringSet(
  seq, fafile)
}
#now load Ribo-seq data, reading in the transcript-space alignments
testbam <- system.file('extdata', 'nchr22.bam', package='Ribostan',
  mustWork=TRUE)


## -----------------------------------------------------------------------------

#now load our annootation, filter out suspect ORFs, and find potential uORFs
chr22_anno <- load_annotation(anno_file, fafile, add_uorfs=TRUE)
rpfs <- get_readgr(testbam, chr22_anno)


## -----------------------------------------------------------------------------

#determine offsets by maximum CDS occupancy
offsets_df <- get_offsets(rpfs, chr22_anno)
#use our offsets to determine p-site locations
psites <- get_psite_gr(rpfs, offsets_df, chr22_anno)

## -----------------------------------------------------------------------------

covgrs = list(sample1=rpfs)
metacodondf <- get_metacodon_profs(covgrs, chr22_anno)
kl_df<-get_kl_df(metacodondf, chr22_anno)
kl_offsets <- select_offsets(kl_df)
kl_offsets%>%dplyr::select(p_offset,length=nreadlen)%>%
  readr::write_tsv('offsets_rustvar.tsv')
#
kl_div_plot <- plot_kl_dv(kl_df, kl_offsets)
#
allcodondt <- export_codon_dts(metacodondf, kl_offsets)
knitr::kable(allcodondt)


## -----------------------------------------------------------------------------
#use our p-sites to filter the annotation 
chr22_anno <- periodicity_filter_uORFs( psites, chr22_anno, remove=TRUE)
#now, use Stan to estimate normalized p-site densities for our data
ritpms <- get_ritpms(psites, chr22_anno)
#and get these at the gene level (ignoring uORFs)
gritpms = gene_level_expr(ritpms, chr22_anno)
#and remove multimapped alignments from the psites
psites <- psites[psites$orf%in%names(chr22_anno$trspacecds)]
psites <- sample_cov_gr(psites, chr22_anno, ritpms)

knitr::kable(tibble::enframe(head(ritpms)))
knitr::kable(head(gritpms))

## -----------------------------------------------------------------------------
#get codon-level occupanies using RUST (at least 1000 should be used 
#for n_genes on  a real run)
rust_codon_occ_df <- get_codon_occs(psites, offsets_df, chr22_anno,
  n_genes=20, method='RUST_glm')
#get predicted mean elongation rates for ORFs
orf_elong = get_orf_elong(chr22_anno, rust_codon_occ_df)
#and at the gene level, ingoring uORFs
gn_elong = gene_level_elong(orf_elong, ritpms, chr22_anno)
knitr::kable(head(gn_elong))


